package fr.inria.gemo.axml.optimax.strategy;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.io.input.FileInputModule;

/**
 * A strategy is  a sequence of 1 or more steps.
 * A step is: a search algorithm, followed by either an atom repetition or an atom.
 * A search algorithm is a sequence of 1 or more algorithm atoms such as D (depth), 
 *   B (breadth), NC (network cost) etc.
 * A repetition may be unbounded (*, which means until none is possible) or bounded (5).
 * An atom is either a disjunction or moves, or a wildcard standing for any move (X),
 * or an atom.
 * An atom is either: 
 * - common sub-expression elimination (CSE)
 * - instantiation (INST)
 * - delegation	(DL)
 * - query composition (C)
 * For example: 
 * (CSE|INST|DL|C)* means do all sequences of moves from this alphabet in any order.
 * (SE)5 means do all sequences of at most 5 common sub-expression eliminations. 
 * (SE)5,(INST)* means do all sequences of at most 5 CSE followed by as many INST 
 *               as you like.
 * (INST)*,(CSE)*,(DL)* means... etc.
 * @author ioana
 *
 */
public class Strategy {

	public ArrayList<Step> steps;

	public Strategy(){
		steps = new ArrayList<Step>();
	}

	public Strategy (String XMLFileName){
		steps = new ArrayList<Step>();
		FileInputModule fim = new FileInputModule();

		org.w3c.dom.Document d = fim.getDomFromXmlFile(XMLFileName);
		NodeList strategyList = d.getChildNodes();
		Node strategyNode = strategyList.item(0);

		// list of steps
		NodeList stepNodes = strategyNode.getChildNodes();
		for (int i = 0; i < stepNodes.getLength(); i ++){
			Node step = stepNodes.item(i);
			// kill everything but the elements
			if (step.getNodeType() != Node.ELEMENT_NODE){
				continue;
			}
			System.out.println("\nUNDERSTOOD STEP:");
			// what does the step have?
			NodeList stepChildren = step.getChildNodes();
			
			LinkedList<SearchAlgorithmAtom> algoAtoms = new LinkedList<SearchAlgorithmAtom>();
			
			// children can be:
			// rewriting rules
			// algorithms
			// cardinality (just 1)
			
			for (int j = 0; j < stepChildren.getLength(); j ++){
				Node stepChild = stepChildren.item(j);
				if (stepChild.getNodeType()!= Node.ELEMENT_NODE){
					continue;
				}
				String sName = stepChild.getLocalName();
				if (sName.compareTo("algo")==0){
					algoAtoms.add(new SearchAlgorithmAtom(stepChild.getTextContent()));
					//System.out.println("AlgoAtom not null");
				}
				else{
					if (sName.compareTo("atom")==0){
						RuleAtom a = new RuleAtom(stepChild.getTextContent());
						AtomStep as = new AtomStep(algoAtoms, a);
						steps.add(as);
					}
					else{
						if (sName.compareTo("repetition")==0){
							System.out.println("Understood a repetition");
							NodeList repetitionExpr = stepChild.getChildNodes();

							// trying to understand the cardinality: looking at the end
							Node tentativeIndicator = null;						
							int k = repetitionExpr.getLength() - 1; 
							for(; k > 0; k --){
								tentativeIndicator = repetitionExpr.item(k);
								if (tentativeIndicator.getNodeType()==Node.ELEMENT_NODE){
									break;
								}
							}
							CardinalityAtom cardIndicator = new CardinalityAtom(tentativeIndicator.getTextContent());
							
							ArrayList<RuleAtom> atoms = getAtomAlternatives(repetitionExpr,
									0, k);
							RepetitionStep rs = new RepetitionStep(algoAtoms, atoms, cardIndicator);
							steps.add(rs);
						}	
						else{
							throw new Error("Illegal element name: " + sName);
						}	
					}
				}
			}
		}
		System.out.println(this.toString());
	}
	
	private static final ArrayList<RuleAtom> getAtomAlternatives(NodeList nl, int iBegin, int iEnd){
		ArrayList res = new ArrayList<RuleAtom>();
		for (int i = iBegin; i < nl.getLength() && i < iEnd; i ++){
			Node n = nl.item(i);
			if (n.getNodeType()!= Node.ELEMENT_NODE){
				continue;
			}
			String s = n.getTextContent();
			if (s.compareTo(Parameters.BREADTH_FIRST)!= 0 &&
					s.compareTo(Parameters.COMPOSITION)!=0 &&
					s.compareTo(Parameters.CSE) !=0 &&
					s.compareTo(Parameters.DELEGATION) != 0 &&
					s.compareTo(Parameters.DEPTH_FIRST) !=0 &&
					s.compareTo(Parameters.GENERALIZATION) != 0 &&
					s.compareTo(Parameters.INSTANTIATION) !=0){
				throw new Error("Wrong text content: " + s);
			}
			res.add(new RuleAtom(n.getTextContent()));
		}
		return res;
	}

	public String toString(){
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < steps.size(); i ++){
			Step s = (Step)steps.get(i);
			s.toString(sb);
			if (i < steps.size() - 1){
				sb.append(", ");
			}
		}
		return new String(sb);
	}

	public static final Strategy defaultStrategy(String algorithm){
		Strategy strat = new Strategy();
		strat.steps = new ArrayList<Step>();
		SearchAlgorithmAtom saa = new SearchAlgorithmAtom(algorithm);
		LinkedList<SearchAlgorithmAtom> lsa = new LinkedList<SearchAlgorithmAtom>();
		lsa.add(saa);
		RuleAtom delegAtom = new RuleAtom(Parameters.DELEGATION);
		RuleAtom cseAtom = new RuleAtom(Parameters.CSE);
		CardinalityAtom cardAtom = new CardinalityAtom(Parameters.UNBOUNDED);
		ArrayList<RuleAtom> ras = new ArrayList<RuleAtom>();
		ras.add(delegAtom);
		ras.add(cseAtom);
		RepetitionStep singleStep = new RepetitionStep(lsa, ras, cardAtom);
		strat.steps.add(singleStep);
		return strat;
	}
	
	public static void main(String[] argv){
		Strategy s = new Strategy(argv[0]);
	}
}
